#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/mman.h>

int memfd = -1;

unsigned long MMU_DEV_BASE = 0x80000000;
unsigned long MMU_MAP_BASE = 0xffffffff;
unsigned long MMU_MAP_SIZE = 0x10000000;

unsigned long do_mmap (unsigned long addr, unsigned long size)
{
    void *mapped_base, *mapped_dev_base; 
    off_t dev_base, map_size;

	dev_base = addr;
	map_size = size;

    memfd = open("/dev/mem", O_RDWR | O_SYNC);
	if (memfd == -1) {
        printf("Can't open /dev/mem.\n");
		return -1;
    }

    printf("/dev/mem opened.\n");
 
    // Map one page of memory into user space such that the device is in that page, but it may not
    // be at the start of the page
 
    mapped_base = mmap(0, map_size, PROT_READ | PROT_WRITE, MAP_SHARED, memfd, dev_base);
	if (mapped_base == (void *) -1) {
        printf("Can't map the memory to user space.\n");
		return -1;
    }

    printf("Memory 0x%08x (size 0x%08x) mapped at address 0x%08x\r\n", addr, size, mapped_base);

	MMU_DEV_BASE = dev_base;
	MMU_MAP_BASE = mapped_base;
	MMU_MAP_SIZE = size;

	return mapped_base;
}

unsigned long do_munmap (unsigned long addr, unsigned long size)
{
    if (munmap(addr, size) == -1) {
        printf("Can't unmap memory from user space.\n");
        return -1;
    }

    close(memfd);
	memfd = -1;

	return 0;
}

unsigned long MMU_ADDR (unsigned long phys)
{
	unsigned long virt = MMU_MAP_BASE + (phys - MMU_DEV_BASE);

	if (virt < MMU_MAP_BASE || virt > MMU_MAP_BASE + MMU_MAP_SIZE)
	{
		fprintf (stderr, "Address out of range: 0x%08x [0x%08x - 0x%08x]\n", virt, MMU_MAP_BASE, MMU_MAP_BASE + MMU_MAP_SIZE);
		exit(1);
	}

	return virt;
}
